import os, random, time

def rsleep():
    time.sleep(random.random())

class base_class(object):
    def send_message(self, to, data, pid = None, token = None):
        bus = self.bus
        if pid is None:
            try:
                pid = self.pid
                assert not pid is None
            except:
                pid = os.getpid()
                self.pid = pid
        if token is None:
            token = random.randint(1000,9999)
        msg = {
            'from': pid
            , 'to': to
            , 'data': data
            , 'token': token
            , 'ttl': 10
        }
        # print(msg)
        bus.put(msg)
        return token
    
    def get_message(self, token = None, pid = None):
        bus = self.bus
        if pid is None:
            try:
                pid = self.pid
                assert not pid is None
            except:
                pid = os.getpid()
                self.pid = pid
        while True:
            try:
                if not token is None:
                    msg = self.msg_object[token]
                    del self.msg_object[token]
                    return msg
            except KeyError:
                pass
            msg = bus.get()
            if msg['to'] == pid:
                if not token is None:
                    _token = msg['token']
                    self.msg_object[_token] = msg
                else:
                    return msg
            else:
                msg['ttl'] -= 1
                if msg['ttl'] == 0:
                    print('msg ttl == 0:',msg)
                bus.put(msg)
                rsleep()


from .worker import base_worker, worker
from .util import base_class
from multiprocessing import Process, Queue, Event
import os

class package_boss(base_worker):
    def __init__(self, func):
        q = Queue()
        self._boss = Process(target = func, args = (q, ))
        self._boss.start()
        self.pid = os.getpid()
        super(package_boss, self).__init__(bus = q, boss = self._boss.pid)
    
    def end(self):
        data = {
            'type':'e',
        }
        self.send_message(self.boss, data)
    
    def join(self):
        self._boss.join()

class boss(base_class):
    def __init__(self, q): 
        self.workers = {}
        self.bus = q
        self.pid = os.getpid()
        self.handler = {}
        self.handler['f'] = self.kill_worker
        self.handler['r'] = self.new_worker
        self.handler['s'] = self.start
        self.handler['e'] = self.end
        self.isend = False

    def _kill_worker(self, worker):
        token = None
        if worker['pid'] is None:
            token = worker['name']
        else:
            token = worker['pid']
        assert not token is None
        worker = self.workers[token]
        if len(worker.sub_workers) > 0:
            for i in worker.sub_workers:
                self._kill_worker(i)
        del self.workers[token]
        worker.terminate()
    
    def kill_worker(self, data):
        self._kill_worker(data['worker'])
        return None

    def new_worker(self, data):
        w = worker(*data['args'], **data['kwargs'], boss = self)
        self.workers[w.name] = w
        data = {
            'name': w.name
            , 'pid' : w.pid
        }
        return data
    
    def start(self, data):
        w = self.workers[data['worker']['name']]
        w.start()
        del self.workers[data['worker']['name']]
        self.workers[w.pid] = w
        data = {
            'name': w.name
            , 'pid' : w.pid
        }
        return data

    def data_handler(self, msg):
        data = msg['data']
        res = self.handler[data['type']](data)
        if not res is None:
            data = res
            # print(data)
            self.send_message(msg['from'], data, token = msg['token'])
    
    def end(self, data):
        for i in workers.values():
            i.kill()
        self.isend = True

    def listen(self):
        msg = None
        while not self.isend:
            msg = self.get_message()
            # print('boss get msg:', msg)
            self.data_handler(msg)


#!/usr/bin/env python
# coding: utf-8

from setuptools import setup

setup(
    name="pboss",
    version='0.0.1',
    author='u2400',
    url="https://pypi.python.org/pypi/pboss",
    download_url="https://pypi.python.org/pypi/pboss",
    description=u'',
    packages=[''],
    install_requires=[],
    entry_points={}
)


from multiprocessing import Process, Queue, Event
from pboss.util import rsleep, base_class
import time, signal, random, os

def n_target(self, func, *args, **kwargs):
    self.pid = os.getpid()
    func(self, *args, **kwargs)

class base_worker(base_class):
    def __init__(self, *, bus, boss):
        self.bus = bus
        self.boss = boss
        self.msg_object = {}
    
    def new_worker(self, *args, **kwargs):
        data = {
            'type': 'r',
            'args': args,
            'kwargs': kwargs
        }
        token = self.send_message(self.boss, data)
        data = self.get_message(token = token)['data']
        # print('new_worker:',data)
        pid = data['pid']
        name = data['name']
        return package_worker(pid, name, bus = self.bus, boss = self.boss)
    
    def start_worker(self, worker):
        assert not((worker.name is None) and (worker.pid is None))
        data = {
            'type': 's',
            'worker': {
                'name': worker.name
                , 'pid': worker.pid
            }
        }
        token = self.send_message(self.boss, data)
        return self.get_message(token)['data']

    def kill_worker(self, worker):
        assert not((worker.name is None) and (worker.pid is None))
        token = None
        data = {
            'type': 'f',
            'worker': {
                'name': worker.name
                , 'pid': worker.pid
            }
        }
        self.send_message(self.boss, data)

class package_worker(base_worker):
    def __init__(self, pid, name, *args, **kwargs):
        super(package_worker, self).__init__(*args, **kwargs)
        self.pid = pid
        self.name = name
    
    def start(self):
        data = self.start_worker(self)
        self.name = data['name']
        self.pid = data['pid']

    def kill(self):
        self.kill_worker(self)

class worker(Process, base_worker):
    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}, *, daemon=None, boss):
        self.sub_workers = []
        self.bus = boss.bus
        self.boss = boss.pid
        pkw = package_worker(None, None, bus = self.bus, boss = self.boss)
        args = (pkw, target) + args
        Process.__init__(self, group, n_target, name, args, kwargs, daemon=daemon)